from .port import INPUT, OUTPUT, INOUT
from .t import In, Out, InOut
from .bit import Bit, _BitKind, _BitType
from .wire import wire

__all__  = ['ClockKind', 'ClockType']
__all__ += ['Clock', 'ClockIn', 'ClockOut']

__all__ += ['ResetKind', 'ResetType']
__all__ += ['Reset', 'ResetIn', 'ResetOut']

__all__ += ['ResetNKind', 'ResetNType']
__all__ += ['ResetN', 'ResetNIn', 'ResetNOut']

__all__ += ['AsyncResetKind', 'AsyncResetType']
__all__ += ['AsyncReset', 'AsyncResetIn', 'AsyncResetOut']

__all__ += ['AsyncResetNKind', 'AsyncResetNType']
__all__ += ['AsyncResetN', 'AsyncResetNIn', 'AsyncResetNOut']

__all__ += ['EnableKind', 'EnableType']
__all__ += ['Enable', 'EnableIn', 'EnableOut']

__all__ += ['ClockInterface', 'ClockTypes']
__all__ += ['wireclock', 'wireclocktype', 'wiredefaultclock']




class ClockKind(_BitKind):
    def __str__(cls):
        if cls.isinput():  return 'In(Clock)'
        if cls.isoutput(): return 'Out(Clock)'
        return 'Clock'

    def qualify(cls, direction):
        if   direction is None:   return Clock
        elif direction == INPUT:  return ClockIn
        elif direction == OUTPUT: return ClockOut
        return cls

    def flip(cls):
        if   cls.isoriented(INPUT):  return ClockOut
        elif cls.isoriented(OUTPUT): return ClockIn
        return cls

class ClockType(_BitType):
    pass

Clock = ClockKind('Clock', (ClockType,), {})
ClockIn = ClockKind('Clock', (ClockType,), dict(direction=INPUT))
ClockOut = ClockKind('Clock', (ClockType,), dict(direction=OUTPUT))

# synchronous reset, active high (i.e. reset when signal is 1)
class ResetKind(_BitKind):
    def __str__(cls):
        if cls.isinput():  return 'In(Reset)'
        if cls.isoutput(): return 'Out(Reset)'
        return 'Reset'

    def qualify(cls, direction):
        if   direction is None:   return Reset
        elif direction == INPUT:  return ResetIn
        elif direction == OUTPUT: return ResetOut
        return cls

    def flip(cls):
        if   cls.isoriented(INPUT):  return ResetOut
        elif cls.isoriented(OUTPUT): return ResetIn
        return cls

class ResetType(_BitType):
    pass

Reset = ResetKind('Reset', (ResetType,), {})
ResetIn = ResetKind('Reset', (ResetType,), dict(direction=INPUT))
ResetOut = ResetKind('Reset', (ResetType,), dict(direction=OUTPUT))

# synchronous reset, active low (i.e. reset when signal is 0)
class ResetNKind(_BitKind):
    def __str__(cls):
        if cls.isinput():  return 'In(ResetN)'
        if cls.isoutput(): return 'Out(ResetN)'
        return 'ResetN'

    def qualify(cls, direction):
        if   direction is None:   return ResetN
        elif direction == INPUT:  return ResetNIn
        elif direction == OUTPUT: return ResetNOut
        return cls

    def flip(cls):
        if   cls.isoriented(INPUT):  return ResetNOut
        elif cls.isoriented(OUTPUT): return ResetNIn
        return cls

class ResetNType(_BitType):
    pass

ResetN = ResetNKind('ResetN', (ResetNType,), {})
ResetNIn = ResetNKind('ResetN', (ResetNType,), dict(direction=INPUT))
ResetNOut = ResetNKind('ResetN', (ResetNType,), dict(direction=OUTPUT))

# asynchronous reset, active high (i.e. reset when signal is 1)
class AsyncResetKind(_BitKind):
    def __str__(cls):
        if cls.isinput():  return 'In(AsyncReset)'
        if cls.isoutput(): return 'Out(AsyncReset)'
        return 'AsyncReset'

    def qualify(cls, direction):
        if   direction is None:   return AsyncReset
        elif direction == INPUT:  return AsyncResetIn
        elif direction == OUTPUT: return AsyncResetOut
        return cls

    def flip(cls):
        if   cls.isoriented(INPUT):  return AsyncResetOut
        elif cls.isoriented(OUTPUT): return AsyncResetIn
        return cls

class AsyncResetType(_BitType):
    pass

AsyncReset = AsyncResetKind('AsyncReset', (AsyncResetType,), {})
AsyncResetIn = AsyncResetKind('AsyncReset', (AsyncResetType,), dict(direction=INPUT))
AsyncResetOut = AsyncResetKind('AsyncReset', (AsyncResetType,), dict(direction=OUTPUT))

# asynchronous reset, active low (i.e. reset when signal is 0)
class AsyncResetNKind(_BitKind):
    def __str__(cls):
        if cls.isinput():  return 'In(AsyncResetN)'
        if cls.isoutput(): return 'Out(AsyncResetN)'
        return 'AsyncResetN'

    def qualify(cls, direction):
        if   direction is None:   return AsyncResetN
        elif direction == INPUT:  return AsyncResetNIn
        elif direction == OUTPUT: return AsyncResetNOut
        return cls

    def flip(cls):
        if   cls.isoriented(INPUT):  return AsyncResetNOut
        elif cls.isoriented(OUTPUT): return AsyncResetNIn
        return cls

class AsyncResetNType(_BitType):
    pass

AsyncResetN = AsyncResetNKind('AsyncResetN', (AsyncResetNType,), {})
AsyncResetNIn = AsyncResetNKind('AsyncResetN', (AsyncResetNType,), dict(direction=INPUT))
AsyncResetNOut = AsyncResetNKind('AsyncResetN', (AsyncResetNType,), dict(direction=OUTPUT))

# Preset
# Clear

class EnableKind(_BitKind):
    def __str__(cls):
        if cls.isinput():  return 'In(Enable)'
        if cls.isoutput(): return 'Out(Enable)'
        return 'Enable'

    def qualify(cls, direction):
        if   direction is None:   return Enable
        elif direction == INPUT:  return EnableIn
        elif direction == OUTPUT: return EnableOut
        return cls

    def flip(cls):
        if   cls.isoriented(INPUT):  return EnableOut
        elif cls.isoriented(OUTPUT): return EnableIn
        return cls

class EnableType(_BitType):
    pass

Enable = EnableKind('Enable', (EnableType,), {})
EnableIn = EnableKind('Enable', (EnableType,), dict(direction=INPUT))
EnableOut = EnableKind('Enable', (EnableType,), dict(direction=OUTPUT))

ClockTypes = (ClockType, ResetType, AsyncResetType, AsyncResetNType, EnableType)


def ClockInterface(has_enable=False, has_reset=False, has_set=False,
                   has_ce=False, has_async_reset=False,
                   has_async_resetn=False):
    args = ['CLK', In(Clock)]
    has_enable |= has_ce
    if has_enable:
        args += ['CE', In(Enable)]
    if has_reset:
        args += ['RESET', In(Reset)]
    if has_async_reset:
        args += ['ASYNCRESET', In(AsyncReset)]
    if has_async_resetn:
        args += ['ASYNCRESETN', In(AsyncResetN)]
    return args


def wireclocktype(defn, inst, clocktype):
    #print('wiring clocks', str(defn), str(inst))
    defnclk = []
    for name, port in defn.interface.ports.items():
         if isinstance(port, clocktype) and port.isoutput():
             #print('defn clock', port)
             defnclk += [port]
    if defnclk:
        defnclk = defnclk[0] # wire first clock
        for name, port in inst.interface.ports.items():
             if isinstance(port, clocktype) and port.isinput() and not port.driven():
                 #print('inst clock', port)
                 wire(defnclk, port)

def wiredefaultclock(defn, inst):
    wireclocktype(defn, inst, ClockType)

def wireclock(define, circuit):
    wireclocktype(define, circuit, ResetType)
    wireclocktype(define, circuit, AsyncResetType)
    wireclocktype(define, circuit, AsyncResetNType)
    wireclocktype(define, circuit, EnableType)



from setuptools import setup
import sys

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='magma-lang',
    version='1.0.24',
    url='https://github.com/phanrahan/magma',
    license='MIT',
    maintainer='Lenny Truong',
    maintainer_email='lenny@cs.stanford.edu',
    description='An embedded DSL for constructing hardware circuits',
    scripts=['bin/magma'],
    packages=[
        "magma",
        "magma.frontend",
        "magma.backend",
        "magma.passes",
        "magma.simulator",
        "magma.syntax",
        "magma.ssa",
        "magma.testing"
    ],
    install_requires=[
        "colorlog",
        "astor",
        "six",
        "mako",
        "pyverilog",
        "numpy",
        "graphviz",
        "coreir>=2.0.*",
        "hwtypes>=1.0.*",
        "ast_tools>=0.0.3",
    ],
    python_requires='>=3.6',
    long_description=long_description,
    long_description_content_type="text/markdown"
)



__COMPILE_DIR = 'normal'


def set_compile_dir(target):
    global __COMPILE_DIR
    assert target in ['normal', 'callee_file_dir']
    __COMPILE_DIR = target


def get_compile_dir():
    return __COMPILE_DIR


__DEBUG_MODE = False


def set_debug_mode(value=True):
    global __DEBUG_MODE
    assert value in {True, False}
    __DEBUG_MODE = value


def get_debug_mode():
    return __DEBUG_MODE



from .ref import Ref, AnonRef, DefnRef, InstRef
from .port import INOUT, INPUT, OUTPUT
from .compatibility import IntegerTypes, StringTypes

__all__  = ['Type', 'Kind']
__all__ += ['In', 'Out', 'InOut', 'Flip']


class Type(object):
    def __init__(self, **kwargs):
        # ref is int, str or tuple
        name = kwargs.get('name', None)
        if name is None or isinstance(name, str):
            #print('creating name ref',name)
            name = AnonRef(name=name)
        #print('using',name)
        #assert isinstance(name, Ref)
        self.name = name

    # subclasses only need to implement one of these methods
    def __eq__(self, rhs):
        return not (self != rhs)
    def __ne__(self, rhs):
        return not (self == rhs)

    __hash__ = object.__hash__

    def __repr__(self):
        return repr(self.name)

    def __str__(self):
        return str(self.name)

    # an instance has an anon name
    def anon(self):
        return self.name.anon()

    # abstract method - must be implemented by subclasses
    @classmethod
    def isoriented(cls, direction):
        pass

    @classmethod
    def isinput(cls):
        return cls.isoriented(INPUT)

    @classmethod
    def isoutput(self):
        return self.isoriented(OUTPUT)

    @classmethod
    def isinout(self):
        return self.isoriented(INOUT)

    @classmethod
    def isbidir(self):
        return False

    @property
    def debug_name(self):
        defn_str = ""
        inst_str = ""
        if isinstance(self.name, DefnRef):
            defn_str = str(self.name.defn.name) + "."
        elif isinstance(self.name, InstRef):
            inst_str = str(self.name.inst.name) + "."
            defn_str = str(self.name.inst.defn.name) + "."
        return f"{defn_str}{inst_str}{str(self)}"

    def __le__(self, other):
        if not self.isoutput():
            self.wire(other)
        else:
            raise TypeError(f"Cannot use <= to assign to output: {self.debug_name} (trying to assign {other.debug_name})")


class Kind(type):
    def __init__(cls, name, bases, dct):
        type.__init__( cls, name, bases, dct)

    # subclasses only need to implement one of these methods
    def __eq__(cls, rhs):
        return not (cls != rhs)
    def __ne__(cls, rhs):
        return not (cls == rhs)

    __hash__ = type.__hash__

    def __repr__(cls):
        return cls.__name__

    def __str__(cls):
        return cls.__name__

    # abstract method - must be implemented by subclasses
    def qualify(cls):
        pass

    # abstract method - must be implemented by subclasses
    def flip(cls):
        pass




def In(T):
    return T.qualify(direction=INPUT)

def Out(T):
    return T.qualify(direction=OUTPUT)

def InOut(T):
    return T.qualify(direction=INOUT)

def Flip(T):
    return T.flip()



from .compatibility import IntegerTypes


__all__ = ['AnonRef', 'InstRef', 'DefnRef', 'ArrayRef', 'TupleRef']


class Ref:
    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return self.qualifiedname()


class AnonRef(Ref):
    def __init__(self, name=""):
        self.name = name

    def __str__(self):
        return str(self.name) if self.name else f"AnonymousValue_{id(self)}"

    def qualifiedname(self, sep='.'):
        return str(self.name) if self.name else f"AnonymousValue_{id(self)}"

    def anon(self):
        return False if self.name else True


class InstRef(Ref):
    def __init__(self, inst, name):
        if not inst:
            raise ValueError(f"Bad inst: {inst}")
        self.inst = inst
        self.name = name

    def qualifiedname(self, sep="."):
        name = self.name
        if isinstance(self.name, IntegerTypes):
            # Hack, Hack, Hack!
            if sep == ".":
                return f"{self.inst.name}[{self.name}]"
        return self.inst.name + sep + str(name)

    def anon(self):
        return False


class DefnRef(Ref):
    def __init__(self, defn, name):
        if not defn:
            raise ValueError(f"Bad defn: {defn}")
        self.defn = defn
        self.name = name

    def qualifiedname(self, sep="."):
        if sep == ".":
            return self.defn.__name__ + sep + self.name
        return self.name

    def anon(self):
        return False


class ArrayRef(Ref):
   def __init__(self, array, index):
       self.array = array
       self.index = index

   def __str__(self):
       return self.qualifiedname()

   def qualifiedname(self, sep="."):
       return f"{self.array.name.qualifiedname(sep=sep)}[{self.index}]"

   def anon(self):
       return self.array.name.anon()


class TupleRef(Ref):
   def __init__(self, tuple, index):
       self.tuple = tuple
       self.index = index

   def __str__(self):
       return self.qualifiedname()

   def qualifiedname(self, sep="."):
       return self.tuple.name.qualifiedname(sep=sep) + sep + str(self.index)

   def anon(self):
       return self.tuple.name.anon()



import inspect
import typing
import ast
from .util import get_ast
import astor
import magma.ast_utils as ast_utils
from magma.debug import debug_info
import functools
import magma as m
from magma.ssa import convert_tree_to_ssa
from magma.config import get_debug_mode
from collections import Counter
import itertools

from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, SymbolTable

class RewriteSelfAttributes(ast.NodeTransformer):
    def __init__(self, initial_value_map):
        self.initial_value_map = initial_value_map
        self.calls_seen = []

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name) and node.value.id == "self":
            if isinstance(node.ctx, ast.Store):
                return ast.Name(f"self_{node.attr}_I", ast.Store())
            else:
                return ast.Name(f"self_{node.attr}_O", ast.Load())
        return node

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute) and \
                isinstance(node.func.value, ast.Name) and \
                node.func.value.id == "self":
            attr = node.func.attr
            assert attr in self.initial_value_map, \
                "Reference to self that was not initialized"
            ret = []
            ports = self.initial_value_map[attr][3].interface.inputs()
            for name, value in zip(ports, node.args):
                ret.append(ast.parse(
                    f"self_{attr}_{name} = {astor.to_source(value).rstrip()}"
                ).body[0])
            self.calls_seen.extend(ret)
            func = astor.to_source(node.func).rstrip()
            outputs = self.initial_value_map[attr][3].interface.outputs()
            if len(outputs) == 1:
                return ast.Name(f"self_{attr}_{outputs[0]}", ast.Load())
            else:
                assert outputs, "Expected module with at least one output"
                return ast.Tuple([ast.Name(f"self_{attr}_{output}",
                                           ast.Load()) for output in outputs],
                                 ast.Load())
        return node

    def visit_If(self, node):
        node.test = self.visit(node.test)
        for attr in ["body", "orelse"]:
            new_body = []
            for item in getattr(node, attr):
                new_body.append(self.visit(item))
            if self.calls_seen:
                new_body = self.calls_seen + new_body
                self.calls_seen = []
            setattr(node, attr, new_body)
        return node


    # def visit(self, node):
    #     node = super().visit(node)
    #     if hasattr(node, 'body'):
    #         new_body = self.calls_seen
    #         self.calls_seen = []
    #         for item in node.body:
    #             if isinstance(item, list):
    #                 new_body.extend(item)
    #             else:
    #                 new_body.append(item)
    #         node.body = new_body
    #     return node


class RewriteReturn(ast.NodeTransformer):
    def __init__(self, initial_value_map):
        self.initial_value_map = initial_value_map

    def visit_Return(self, node):
        elts = []
        for name, (value, type_, eval_type, eval_value) in self.initial_value_map.items():
            if isinstance(eval_type, m.Kind):
                elts.append(ast.Name(f"self_{name}_I", ast.Load()))
            else:
                for port in eval_value.interface.inputs():
                    if isinstance(port, (m.ClockType, m.AsyncResetType)):
                        continue
                    elts.append(ast.Name(f"self_{name}_{port}", ast.Load()))
        if isinstance(node.value, ast.Tuple):
            elts.extend(node.value.elts)
        else:
            elts.append(node.value)
        node.value = ast.Tuple(elts, ast.Load())
        return node


def get_initial_value_map(init_func, defn_env):
    """
    Parses __init__ funciton of the form

        def __init__(self):
            self.x = m.bits(0, 2)
            self.y = m.bits(0, 4)

    Returns a map from instance attributes to their initial values
    For the above example,

        {'x': <_ast.Call object at 0x109c17fd0>,
         'y': <_ast.Call object at 0x109c17470>}

    TODO: Should we allow arbitrary code in the definition? For now, the user
    can write any meta code outside the class definition, which makes the
    design of this simpler
    """
    initial_value_map = {}
    init_def = get_ast(init_func).body[0]
    # init_def = ExecuteEscapedPythonExpressions(defn_env).visit(init_def)
    init_def = SpecializeConstantInts(defn_env).visit(init_def)
    for stmt in init_def.body:
        # We only support basic assignments of the form
        #     self.x: m.Bits(2) = m.bits(0, 2)
        assert isinstance(stmt, ast.AnnAssign)
        assert isinstance(stmt.target, ast.Attribute)
        assert isinstance(stmt.target.value, ast.Name)
        assert stmt.target.value.id == "self"
        # TODO: Should we deal with multiple assignments? For now we take the
        # last one
        m.DefineCircuit("tmp")
        eval_type = eval(astor.to_source(stmt.annotation).rstrip(), defn_env)
        eval_value = eval(astor.to_source(stmt.value).rstrip(), defn_env)
        m.EndCircuit()
        initial_value_map[stmt.target.attr] = (stmt.value, stmt.annotation,
                                               eval_type, eval_value)
    return initial_value_map


def get_io(call_def):
    """
    Parses a __call__ method of the form

        def __call__(self, I):
            O = self.y
            self.y = self.x
            self.x = I
            return O

    Returns a tuple
    [0]: the list of tuples containing the name and type of each input argument
         excluding `self`
    [1]: the output type
    """
    # Only support basic args for now
    assert not call_def.args.vararg
    assert not call_def.args.kwonlyargs
    assert not call_def.args.kwarg
    # assert not call_def.args.defaults
    assert not call_def.args.kw_defaults

    # Skips self
    assert call_def.args.args[0].arg == "self"
    inputs = [(arg.arg, arg.annotation) for arg in call_def.args.args[1:]]
    returns = call_def.returns
    return inputs, call_def.returns


circuit_definition_template = """
from magma import In, Out, Bit

def make_{circuit_name}(combinational):
    class {circuit_name}(m.Circuit):
        IO = {io_list}

        @classmethod
        def definition(io):
            {register_instances}
            @combinational
            def {circuit_name}_comb({circuit_combinational_args}) -> ({circuit_combinational_output_type}):
                {circuit_combinational_body}
            comb_out = {circuit_name}_comb({circuit_combinational_call_args})
            {comb_out_wiring}
    return {circuit_name}
"""


def gen_array_str(eval_type, eval_value, async_reset):
    arr = [gen_reg_inst_str(eval_type.T, eval_value[i], async_reset) for i
           in range(len(eval_type))]
    return f"m.join([{', '.join(arr)}])"


def gen_reg_inst_str(eval_type, eval_value, async_reset):
    if isinstance(eval_type, m._BitKind):
        n = None
        if isinstance(eval_value, (bool, int)):
            init = bool(eval_value)
        else:
            assert eval_value.name.name in ["GND", "VCC"], eval_value.name
            init = 0 if eval_value.name.name == "GND" else 1
    elif isinstance(eval_type, m.ArrayKind) and \
            isinstance(eval_type.T, m._BitKind):
        n = len(eval_type)
        init = int(eval_value)
    elif isinstance(eval_type, m.ArrayKind):
        return f"{gen_array_str(eval_type, eval_value, async_reset)}"
    else:
        raise NotADirectoryError((eval_type))
    return f"DefineRegister({n}, init={init}, has_async_reset={async_reset})()"


def gen_register_instances(initial_value_map, async_reset):
    """
    Generates a sequence of statements to instance a set of registers from
    `initial_value_map`.

    For example,

        initial_value_map = {'x': <_ast.Call object at 0x109c17fd0>,
                             'y': <_ast.Call object at 0x109c17470>}

    will generate

        ['x = Register(2, init=0)', 'y = Register(4, init=0)']

    """
    register_instances = []
    for name, (value, type_, eval_type, eval_value) in initial_value_map.items():
        if isinstance(eval_type, m.Kind):
            reg_inst_str = gen_reg_inst_str(eval_type, eval_value,
                                            async_reset)
            register_instances.append(f"{name} = {reg_inst_str}")
        else:
            value = astor.to_source(value).rstrip()
            register_instances.append(f"{name} = {value}")
    return register_instances


def gen_io_list(inputs, output_type, async_reset):
    io_list = []
    for name, type_ in inputs:
        type_ = astor.to_source(type_).rstrip()
        io_list.append(f"\"{name}\", m.In({type_})")
    io_list.append(f"\"CLK\", m.In(m.Clock)")
    if async_reset:
        io_list.append(f"\"ASYNCRESET\", m.In(m.AsyncReset)")
    if isinstance(output_type, ast.Tuple):
        outputs = []
        for i, elem in enumerate(output_type.elts):
            output_type_str = astor.to_source(elem).rstrip()
            io_list.append(f"\"O{i}\", m.Out({output_type_str})")
    else:
        output_type_str = astor.to_source(output_type).rstrip()
        io_list.append(f"\"O\", m.Out({output_type_str})")
    return '[' + ', '.join(io_list) + ']'


class EscapedExpression(ast.AST):
    def __init__(self, value, orig):
        self.value = value
        self.orig = orig


class ExecuteEscapedPythonExpressions(ast.NodeTransformer):
    def __init__(self, defn_env):
        self.defn_env = defn_env

    def visit_List(self, node):
        assert len(node.elts) == 1, "Expected only list literatals that " \
            "contain a single expression"
        # FIXME: Hack to prevent instancing logic from polluting currently open
        # definition
        m.DefineCircuit("tmp")
        result = eval(astor.to_source(node.elts[0]).rstrip(), self.defn_env)
        m.EndCircuit()
        return EscapedExpression(result, node)


class SpecializeConstantInts(ast.NodeTransformer):
    def __init__(self, defn_env):
        self.defn_env = defn_env

    def visit_Name(self, node):
        if node.id in self.defn_env:
            value = self.defn_env[node.id]
            if isinstance(value, int):
                return ast.Num(value)
        return node


def _sequential(
        defn_env: dict,
        async_reset: bool,
        cls,
        combinational_decorator: typing.Callable):
    # if not inspect.isclass(cls):
    #     raise ValueError("sequential decorator only works with classes")

    initial_value_map = get_initial_value_map(cls.__init__, defn_env)

    call_def = get_ast(cls.__call__).body[0]
    inputs, output_type = get_io(call_def)
    io_list = gen_io_list(inputs, output_type, async_reset)

    circuit_combinational_output_type = []
    circuit_combinational_args = []
    circuit_combinational_call_args = []
    comb_out_wiring = []
    for name, type_ in inputs:
        type_ = astor.to_source(type_).rstrip()
        circuit_combinational_args.append(f"{name}: {type_}")
        circuit_combinational_call_args.append(f"io.{name}")

    comb_out_count = 0
    for name, (value, type_, eval_type, eval_value) in initial_value_map.items():
        if isinstance(eval_type, m.Kind):
            type_ = astor.to_source(type_).rstrip()
            circuit_combinational_args.append(f"self_{name}_O: {type_}")
            circuit_combinational_call_args.append(f"{name}")
            circuit_combinational_output_type.append(f"{type_}")
            comb_out_wiring.append(f"{name}.I <= comb_out[{comb_out_count}]\n")
            comb_out_count += 1
        else:
            for key, value in eval_value.interface.ports.items():
                if isinstance(value, (m.ClockType, m.AsyncResetType)):
                    continue
                type_ = repr(type(value))
                if value.isoutput():
                    circuit_combinational_args.append(f"self_{name}_{value}: m.{type_}")
                    circuit_combinational_call_args.append(f"{name}.{value}")
                if value.isinput():
                    circuit_combinational_output_type.append(f"m.{type_}")
                    comb_out_wiring.append(f"{name}.{value} <= comb_out[{comb_out_count}]\n")
                    comb_out_count += 1

    circuit_combinational_args = ', '.join(circuit_combinational_args)
    circuit_combinational_call_args = ', '.join(circuit_combinational_call_args)

    if isinstance(output_type, ast.Tuple):
        output_types = []
        for i, elem in enumerate(output_type.elts):
            circuit_combinational_output_type.append(astor.to_source(elem).rstrip())
            comb_out_wiring.append(f"io.O{i} <= comb_out[{comb_out_count + i}]\n")
    else:
        output_type_str = astor.to_source(output_type).rstrip()
        circuit_combinational_output_type.append(output_type_str)
        comb_out_wiring.append(f"io.O <= comb_out[{comb_out_count}]\n")

    tab = 4 * ' '
    comb_out_wiring = (3 * tab).join(comb_out_wiring)
    circuit_combinational_output_type = ', '.join(circuit_combinational_output_type)
    circuit_combinational_body = []
    for stmt in call_def.body:
        rewriter = RewriteSelfAttributes(initial_value_map)
        stmt = rewriter.visit(stmt)
        code = [stmt]
        if rewriter.calls_seen:
            code = rewriter.calls_seen + code
        stmt = RewriteReturn(initial_value_map).visit(stmt)
        for stmt in code:
            for line in astor.to_source(stmt).rstrip().splitlines():
                circuit_combinational_body.append(line)

    circuit_combinational_body = ('\n' + 4*tab).join(circuit_combinational_body)
    register_instances = gen_register_instances(initial_value_map, async_reset)
    register_instances = ('\n' + 3*tab).join(register_instances)

    circuit_definition_str = circuit_definition_template.format(
        circuit_name=cls.__name__,
        io_list=io_list,
        register_instances=register_instances,
        circuit_combinational_args=circuit_combinational_args,
        circuit_combinational_output_type=circuit_combinational_output_type,
        circuit_combinational_body=circuit_combinational_body,
        circuit_combinational_call_args=circuit_combinational_call_args,
        comb_out_wiring=comb_out_wiring
    )
    tree = ast.parse(circuit_definition_str)
    if "DefineRegister" not in defn_env:
        tree = ast.Module([
            ast.parse("from mantle import DefineRegister").body[0],
        ] + tree.body)

    circuit_def_constructor = ast_utils.compile_function_to_file(tree, 'make_' + cls.__name__, defn_env)
    circuit_def = circuit_def_constructor(combinational_decorator)

    if get_debug_mode() and getattr(circuit_def, "debug_info", False):
        circuit_def.debug_info = debug_info(circuit_def.debug_info.filename,
                                            circuit_def.debug_info.lineno,
                                            inspect.getmodule(cls))

    return circuit_def

def sequential(
        cls=None,
        async_reset=None,
        *,
        decorators: typing.Optional[typing.Sequence[typing.Callable]] = None,
        env: SymbolTable = None):

    exec(_SKIP_FRAME_DEBUG_STMT)
    if async_reset is not None or decorators is not None:
        assert async_reset is None or isinstance(async_reset, bool)
        assert cls is None
        if decorators is None:
            decorators = ()

        def wrapped(cls):
            exec(_SKIP_FRAME_DEBUG_STMT)
            nonlocal decorators
            decorators = list(itertools.chain(decorators, [wrapped]))
            wrapped_sequential = ast_utils.inspect_enclosing_env(
                    _sequential,
                    decorators=decorators,
                    st=env
            )
            combinational = m.circuit.combinational(decorators=decorators, env=env)
            return wrapped_sequential(async_reset, cls, combinational)
        return wrapped
    else:
        assert cls is not None
        wrapped_sequential = ast_utils.inspect_enclosing_env(
                _sequential,
                decorators=[sequential],
                st=env
        )
        combinational = m.circuit.combinational(decorators=[sequential], env=env)
        return wrapped_sequential(True, cls, combinational)



import dataclasses as dc
from hwtypes.adt_meta import ProductMeta
import magma as m
from collections import OrderedDict


class MagmaProductMeta(ProductMeta):
    @classmethod
    def _from_fields(mcs, fields, name, bases, ns, **kwargs):
        assert fields
        return m.Tuple(**fields)


class Product(metaclass=MagmaProductMeta):
    pass
